
class _Attachments:
    """An _attachment_ object represents any file attached to a task in Asana,
    whether it's an uploaded file or one associated via a third-party service
    such as Dropbox or Google Drive.
    """

    def __init__(self, client=None):
        self.client = client
  
    def find_by_id(self, attachment, params={}, **options): 
        """Returns the full record for a single attachment.

        Parameters
        ----------
        attachment : {Id} Globally unique identifier for the attachment.
        [params] : {Object} Parameters for the request
        """
        path = "/attachments/%s" % (attachment)
        return self.client.get(path, params, **options)
        
    def find_by_task(self, task, params={}, **options): 
        """Returns the compact records for all attachments on the task.

        Parameters
        ----------
        task : {Id} Globally unique identifier for the task.
        [params] : {Object} Parameters for the request
        """
        path = "/tasks/%s/attachments" % (task)
        return self.client.get_collection(path, params, **options)
        




class _Webhooks:
    """Webhooks allow an application to be notified of changes. This is in addition
    to the ability to fetch those changes directly as
    [Events](/developers/api-reference/events) - in fact, Webhooks are just a way
    to receive Events via HTTP POST at the time they occur instead of polling for
    them. For services accessible via HTTP this is often vastly more convenient,
    and if events are not too frequent can be significantly more efficient.
    
    In both cases, however, changes are represented as Event objects - refer to
    the [Events documentation](/developers/api-reference/events) for more
    information on what data these events contain.
    
    **NOTE:** While Webhooks send arrays of Event objects to their target, the
    Event objects themselves contain *only IDs*, rather than the actual resource
    they are referencing. So while a normal event you receive via GET /events
    would look like this:
    
        {\
          "resource": {\
            "id": 1337,\
            "name": "My Task"\
          },\
          "parent": null,\
          "created_at": "2013-08-21T18:20:37.972Z",\
          "user": {\
            "id": 1123,\
            "name": "Tom Bizarro"\
          },\
          "action": "changed",\
          "type": "task"\
        }
    
    In a Webhook payload you would instead receive this:
    
        {\
          "resource": 1337,\
          "parent": null,\
          "created_at": "2013-08-21T18:20:37.972Z",\
          "user": 1123,\
          "action": "changed",\
          "type": "task"\
        }
    
    Webhooks themselves contain only the information necessary to deliver the
    events to the desired target as they are generated.
    """

    def __init__(self, client=None):
        self.client = client
  
    def create(self, params={}, **options): 
        """Establishing a webhook is a two-part process. First, a simple HTTP POST
        similar to any other resource creation. Since you could have multiple
        webhooks we recommend specifying a unique local id for each target.
        
        Next comes the confirmation handshake. When a webhook is created, we will
        send a test POST to the `target` with an `X-Hook-Secret` header as
        described in the
        [Resthooks Security documentation](http://resthooks.org/docs/security/).
        The target must respond with a `200 OK` and a matching `X-Hook-Secret`
        header to confirm that this webhook subscription is indeed expected.
        
        If you do not acknowledge the webhook's confirmation handshake it will
        fail to setup, and you will receive an error in response to your attempt
        to create it. This means you need to be able to receive and complete the
        webhook *while* the POST request is in-flight.

        Parameters
        ----------
        resource : {Id} A resource ID to subscribe to. The resource can be a task or project.
        target : {String} The URL to receive the HTTP POST.
        [data] : {Object} Data for the request
        """
        return self.client.post("/webhooks", params, **options)
        
    def get_all(self, params={}, **options): 
        """Returns the compact representation of all webhooks your app has
        registered for the authenticated user in the given workspace.

        Parameters
        ----------
        workspace : {Id} The workspace to query for webhooks in.
        [params] : {Object} Parameters for the request
          - [resource] : {Id} Only return webhooks for the given resource.
        """
        return self.client.get_collection("/webhooks", params, **options)
        
    def get_by_id(self, webhook, params={}, **options): 
        """Returns the full record for the given webhook.

        Parameters
        ----------
        webhook : {Id} The webhook to get.
        [params] : {Object} Parameters for the request
        """
        path = "/webhooks/%s" % (webhook)
        return self.client.get(path, params, **options)
        
    def delete_by_id(self, webhook, params={}, **options): 
        """This method permanently removes a webhook. Note that it may be possible
        to receive a request that was already in flight after deleting the
        webhook, but no further requests will be issued.

        Parameters
        ----------
        webhook : {Id} The webhook to delete.
        """
        path = "/webhooks/%s" % (webhook)
        return self.client.delete(path, params, **options)
        




class _Projects:
    """A _project_ represents a prioritized list of tasks in Asana or a board with
    columns of tasks represented as cards. It exists in a single workspace or
    organization and is accessible to a subset of users in that workspace or
    organization, depending on its permissions.
    
    Projects in organizations are shared with a single team. You cannot currently
    change the team of a project via the API. Non-organization workspaces do not
    have teams and so you should not specify the team of project in a regular
    workspace.
    """

    def __init__(self, client=None):
        self.client = client
  
    def create(self, params={}, **options): 
        """Creates a new project in a workspace or team.
        
        Every project is required to be created in a specific workspace or
        organization, and this cannot be changed once set. Note that you can use
        the `workspace` parameter regardless of whether or not it is an
        organization.
        
        If the workspace for your project _is_ an organization, you must also
        supply a `team` to share the project with.
        
        Returns the full record of the newly created project.

        Parameters
        ----------
        [data] : {Object} Data for the request
          - workspace : {Id} The workspace or organization to create the project in.
          - [team] : {Id} If creating in an organization, the specific team to create the
          project in.
        """
        return self.client.post("/projects", params, **options)
        
    def create_in_workspace(self, workspace, params={}, **options): 
        """If the workspace for your project _is_ an organization, you must also
        supply a `team` to share the project with.
        
        Returns the full record of the newly created project.

        Parameters
        ----------
        workspace : {Id} The workspace or organization to create the project in.
        [data] : {Object} Data for the request
        """
        path = "/workspaces/%s/projects" % (workspace)
        return self.client.post(path, params, **options)
        
    def create_in_team(self, team, params={}, **options): 
        """Creates a project shared with the given team.
        
        Returns the full record of the newly created project.

        Parameters
        ----------
        team : {Id} The team to create the project in.
        [data] : {Object} Data for the request
        """
        path = "/teams/%s/projects" % (team)
        return self.client.post(path, params, **options)
        
    def find_by_id(self, project, params={}, **options): 
        """Returns the complete project record for a single project.

        Parameters
        ----------
        project : {Id} The project to get.
        [params] : {Object} Parameters for the request
        """
        path = "/projects/%s" % (project)
        return self.client.get(path, params, **options)
        
    def update(self, project, params={}, **options): 
        """A specific, existing project can be updated by making a PUT request on the
        URL for that project. Only the fields provided in the `data` block will be
        updated; any unspecified fields will remain unchanged.
        
        When using this method, it is best to specify only those fields you wish
        to change, or else you may overwrite changes made by another user since
        you last retrieved the task.
        
        Returns the complete updated project record.

        Parameters
        ----------
        project : {Id} The project to update.
        [data] : {Object} Data for the request
        """
        path = "/projects/%s" % (project)
        return self.client.put(path, params, **options)
        
    def delete(self, project, params={}, **options): 
        """A specific, existing project can be deleted by making a DELETE request
        on the URL for that project.
        
        Returns an empty data record.

        Parameters
        ----------
        project : {Id} The project to delete.
        """
        path = "/projects/%s" % (project)
        return self.client.delete(path, params, **options)
        
    def find_all(self, params={}, **options): 
        """Returns the compact project records for some filtered set of projects.
        Use one or more of the parameters provided to filter the projects returned.

        Parameters
        ----------
        [params] : {Object} Parameters for the request
          - [workspace] : {Id} The workspace or organization to filter projects on.
          - [team] : {Id} The team to filter projects on.
          - [archived] : {Boolean} Only return projects whose `archived` field takes on the value of
          this parameter.
        """
        return self.client.get_collection("/projects", params, **options)
        
    def find_by_workspace(self, workspace, params={}, **options): 
        """Returns the compact project records for all projects in the workspace.

        Parameters
        ----------
        workspace : {Id} The workspace or organization to find projects in.
        [params] : {Object} Parameters for the request
          - [archived] : {Boolean} Only return projects whose `archived` field takes on the value of
          this parameter.
        """
        path = "/workspaces/%s/projects" % (workspace)
        return self.client.get_collection(path, params, **options)
        
    def find_by_team(self, team, params={}, **options): 
        """Returns the compact project records for all projects in the team.

        Parameters
        ----------
        team : {Id} The team to find projects in.
        [params] : {Object} Parameters for the request
          - [archived] : {Boolean} Only return projects whose `archived` field takes on the value of
          this parameter.
        """
        path = "/teams/%s/projects" % (team)
        return self.client.get_collection(path, params, **options)
        
    def tasks(self, project, params={}, **options): 
        """Returns the compact task records for all tasks within the given project,
        ordered by their priority within the project. Tasks can exist in more than one project at a time.

        Parameters
        ----------
        project : {Id} The project in which to search for tasks.
        [params] : {Object} Parameters for the request
        """
        path = "/projects/%s/tasks" % (project)
        return self.client.get_collection(path, params, **options)
        
    def add_followers(self, project, params={}, **options): 
        """Adds the specified list of users as followers to the project. Followers are a subset of members, therefore if
        the users are not already members of the project they will also become members as a result of this operation.
        Returns the updated project record.

        Parameters
        ----------
        project : {Id} The project to add followers to.
        [data] : {Object} Data for the request
          - followers : {Array} An array of followers to add to the project.
        """
        path = "/projects/%s/addFollowers" % (project)
        return self.client.post(path, params, **options)
        
    def remove_followers(self, project, params={}, **options): 
        """Removes the specified list of users from following the project, this will not affect project membership status.
        Returns the updated project record.

        Parameters
        ----------
        project : {Id} The project to remove followers from.
        [data] : {Object} Data for the request
          - followers : {Array} An array of followers to remove from the project.
        """
        path = "/projects/%s/removeFollowers" % (project)
        return self.client.post(path, params, **options)
        
    def add_members(self, project, params={}, **options): 
        """Adds the specified list of users as members of the project. Returns the updated project record.

        Parameters
        ----------
        project : {Id} The project to add members to.
        [data] : {Object} Data for the request
          - members : {Array} An array of members to add to the project.
        """
        path = "/projects/%s/addMembers" % (project)
        return self.client.post(path, params, **options)
        
    def remove_members(self, project, params={}, **options): 
        """Removes the specified list of members from the project. Returns the updated project record.

        Parameters
        ----------
        project : {Id} The project to remove members from.
        [data] : {Object} Data for the request
          - members : {Array} An array of members to remove from the project.
        """
        path = "/projects/%s/removeMembers" % (project)
        return self.client.post(path, params, **options)
        
    def add_custom_field_setting(self, project, params={}, **options): 
        """Create a new custom field setting on the project.

        Parameters
        ----------
        project : {Id} The project to associate the custom field with
        [data] : {Object} Data for the request
          - custom_field : {Id} The id of the custom field to associate with this project.
          - [is_important] : {Boolean} Whether this field should be considered important to this project.
          - [insert_before] : {Id} An id of a Custom Field Settings on this project, before which the new Custom Field Settings will be added.
          `insert_before` and `insert_after` parameters cannot both be specified.
          - [insert_after] : {Id} An id of a Custom Field Settings on this project, after which the new Custom Field Settings will be added.
          `insert_before` and `insert_after` parameters cannot both be specified.
        """
        path = "/projects/%s/addCustomFieldSetting" % (project)
        return self.client.post(path, params, **options)
        
    def remove_custom_field_setting(self, project, params={}, **options): 
        """Remove a custom field setting on the project.

        Parameters
        ----------
        project : {Id} The project to associate the custom field with
        [data] : {Object} Data for the request
          - [custom_field] : {Id} The id of the custom field to remove from this project.
        """
        path = "/projects/%s/removeCustomFieldSetting" % (project)
        return self.client.post(path, params, **options)
        




from .gen.events import _Events
from ..error import InvalidTokenError
from ..page_iterator import EventsPageIterator

class Events(_Events):
    """Events resource"""

    def get(self, params, **options):
        """Dispatches a GET request to /events of the API to get a set of recent changes to a resource."""
        options = self.client._merge_options({ 'full_payload': True })
        return self.client.get('/events', params, **options)

    def get_next(self, query, **options):
        """Returns a tuple containing the next page of events and a sync token for the given query (and optional 'sync' token)"""
        iterator = EventsPageIterator(self.client, '/events', query, options)
        result = iterator.next()
        return (result, iterator.sync)

    def get_iterator(self, query, **options):
        """Returns an event iterator for the given query (and optional 'sync' token)"""
        return EventsPageIterator(self.client, '/events', query, options).items()




class _Users:
    """A _user_ object represents an account in Asana that can be given access to
    various workspaces, projects, and tasks.
    
    Like other objects in the system, users are referred to by numerical IDs.
    However, the special string identifier `me` can be used anywhere
    a user ID is accepted, to refer to the current authenticated user.
    """

    def __init__(self, client=None):
        self.client = client
  
    def me(self, params={}, **options): 
        """Returns the full user record for the currently authenticated user.

        Parameters
        ----------
        [params] : {Object} Parameters for the request
        """
        return self.client.get("/users/me", params, **options)
        
    def find_by_id(self, user, params={}, **options): 
        """Returns the full user record for the single user with the provided ID.

        Parameters
        ----------
        user : {String} An identifier for the user. Can be one of an email address,
        the globally unique identifier for the user, or the keyword `me`
        to indicate the current user making the request.
        [params] : {Object} Parameters for the request
        """
        path = "/users/%s" % (user)
        return self.client.get(path, params, **options)
        
    def find_by_workspace(self, workspace, params={}, **options): 
        """Returns the user records for all users in the specified workspace or
        organization.

        Parameters
        ----------
        workspace : {Id} The workspace in which to get users.
        [params] : {Object} Parameters for the request
        """
        path = "/workspaces/%s/users" % (workspace)
        return self.client.get_collection(path, params, **options)
        
    def find_all(self, params={}, **options): 
        """Returns the user records for all users in all workspaces and organizations
        accessible to the authenticated user. Accepts an optional workspace ID
        parameter.

        Parameters
        ----------
        [params] : {Object} Parameters for the request
          - [workspace] : {Id} The workspace or organization to filter users on.
        """
        return self.client.get_collection("/users", params, **options)
        




class _CustomFields:
    """Custom Fields store the metadata that is used in order to add user-specified
    information to tasks in Asana. Be sure to reference the [Custom
    Fields](/developers/documentation/getting-started/custom-fields) developer
    documentation for more information about how custom fields relate to various
    resources in Asana.
    """

    def __init__(self, client=None):
        self.client = client
  
    def create(self, params={}, **options): 
        """Creates a new custom field in a workspace. Every custom field is required to be created in a specific workspace, and this workspace cannot be changed once set.
        
        A custom field's `name` must be unique within a workspace and not conflict with names of existing task properties such as 'Due Date' or 'Assignee'. A custom field's `type` must be one of  'text', 'enum', or 'number'.
        
        Returns the full record of the newly created custom field.

        Parameters
        ----------
        [data] : {Object} Data for the request
          - workspace : {Id} The workspace to create a custom field in.
          - type : {String} The type of the custom field.
          - name : {String} The name of the custom field.
          - [description] : {String} The description of the custom field.
          - [precision] : {Integer} The number of decimal places for the numerical values. Required if the custom field is of type 'number'.
          - [enum_options] : {String} The discrete values the custom field can assume. Required if the custom field is of type 'enum'.
        """
        return self.client.post("/custom_fields", params, **options)
        
    def find_by_id(self, custom_field, params={}, **options): 
        """Returns the complete definition of a custom field's metadata.

        Parameters
        ----------
        custom_field : {Id} Globally unique identifier for the custom field.
        [params] : {Object} Parameters for the request
        """
        path = "/custom_fields/%s" % (custom_field)
        return self.client.get(path, params, **options)
        
    def find_by_workspace(self, workspace, params={}, **options): 
        """Returns a list of the compact representation of all of the custom fields in a workspace.

        Parameters
        ----------
        workspace : {Id} The workspace or organization to find custom field definitions in.
        [params] : {Object} Parameters for the request
        """
        path = "/workspaces/%s/custom_fields" % (workspace)
        return self.client.get_collection(path, params, **options)
        
    def update(self, custom_field, params={}, **options): 
        """A specific, existing custom field can be updated by making a PUT request on the URL for that custom field. Only the fields provided in the `data` block will be updated; any unspecified fields will remain unchanged
        
        When using this method, it is best to specify only those fields you wish to change, or else you may overwrite changes made by another user since you last retrieved the custom field.
        
        A custom field's `type` cannot be updated.
        
        An enum custom field's `enum_options` cannot be updated with this endpoint. Instead see "Work With Enum Options" for information on how to update `enum_options`.
        
        Returns the complete updated custom field record.

        Parameters
        ----------
        custom_field : {Id} Globally unique identifier for the custom field.
        [data] : {Object} Data for the request
        """
        path = "/custom_fields/%s" % (custom_field)
        return self.client.put(path, params, **options)
        
    def delete(self, custom_field, params={}, **options): 
        """A specific, existing custom field can be deleted by making a DELETE request on the URL for that custom field.
        
        Returns an empty data record.

        Parameters
        ----------
        custom_field : {Id} Globally unique identifier for the custom field.
        """
        path = "/custom_fields/%s" % (custom_field)
        return self.client.delete(path, params, **options)
        
    def create_enum_option(self, custom_field, params={}, **options): 
        """Creates an enum option and adds it to this custom field's list of enum options. A custom field can have at most 50 enum options (including disabled options). By default new enum options are inserted at the end of a custom field's list.
        
        Returns the full record of the newly created enum option.

        Parameters
        ----------
        custom_field : {Id} Globally unique identifier for the custom field.
        [data] : {Object} Data for the request
          - name : {String} The name of the enum option.
          - [color] : {String} The color of the enum option. Defaults to 'none'.
          - [insert_before] : {Id} An existing enum option within this custom field before which the new enum option should be inserted. Cannot be provided together with after_enum_option.
          - [insert_after] : {Id} An existing enum option within this custom field after which the new enum option should be inserted. Cannot be provided together with before_enum_option.
        """
        path = "/custom_fields/%s/enum_options" % (custom_field)
        return self.client.post(path, params, **options)
        
    def update_enum_option(self, enum_option, params={}, **options): 
        """Updates an existing enum option. Enum custom fields require at least one enabled enum option.
        
        Returns the full record of the updated enum option.

        Parameters
        ----------
        enum_option : {Id} Globally unique identifier for the enum option.
        [data] : {Object} Data for the request
          - name : {String} The name of the enum option.
          - [color] : {String} The color of the enum option. Defaults to 'none'.
          - [enabled] : {Boolean} Whether or not the enum option is a selectable value for the custom field.
        """
        path = "/enum_options/%s" % (enum_option)
        return self.client.put(path, params, **options)
        
    def insert_enum_option(self, custom_field, params={}, **options): 
        """Moves a particular enum option to be either before or after another specified enum option in the custom field.

        Parameters
        ----------
        custom_field : {Id} Globally unique identifier for the custom field.
        [data] : {Object} Data for the request
          - enum_option : {Id} The ID of the enum option to relocate.
          - name : {String} The name of the enum option.
          - [color] : {String} The color of the enum option. Defaults to 'none'.
          - [before_enum_option] : {Id} An existing enum option within this custom field before which the new enum option should be inserted. Cannot be provided together with after_enum_option.
          - [after_enum_option] : {Id} An existing enum option within this custom field after which the new enum option should be inserted. Cannot be provided together with before_enum_option.
        """
        path = "/custom_fields/%s/enum_options/insert" % (custom_field)
        return self.client.post(path, params, **options)
        




class _Workspaces:
    """A _workspace_ is the highest-level organizational unit in Asana. All projects
    and tasks have an associated workspace.
    
    An _organization_ is a special kind of workspace that represents a company.
    In an organization, you can group your projects into teams. You can read
    more about how organizations work on the Asana Guide.
    To tell if your workspace is an organization or not, check its
    `is_organization` property.
    
    Over time, we intend to migrate most workspaces into organizations and to
    release more organization-specific functionality. We may eventually deprecate
    using workspace-based APIs for organizations. Currently, and until after
    some reasonable grace period following any further announcements, you can
    still reference organizations in any `workspace` parameter.
    """

    def __init__(self, client=None):
        self.client = client
  
    def find_by_id(self, workspace, params={}, **options): 
        """Returns the full workspace record for a single workspace.

        Parameters
        ----------
        workspace : {Id} Globally unique identifier for the workspace or organization.
        [params] : {Object} Parameters for the request
        """
        path = "/workspaces/%s" % (workspace)
        return self.client.get(path, params, **options)
        
    def find_all(self, params={}, **options): 
        """Returns the compact records for all workspaces visible to the authorized user.

        Parameters
        ----------
        [params] : {Object} Parameters for the request
        """
        return self.client.get_collection("/workspaces", params, **options)
        
    def update(self, workspace, params={}, **options): 
        """A specific, existing workspace can be updated by making a PUT request on
        the URL for that workspace. Only the fields provided in the data block
        will be updated; any unspecified fields will remain unchanged.
        
        Currently the only field that can be modified for a workspace is its `name`.
        
        Returns the complete, updated workspace record.

        Parameters
        ----------
        workspace : {Id} The workspace to update.
        [data] : {Object} Data for the request
        """
        path = "/workspaces/%s" % (workspace)
        return self.client.put(path, params, **options)
        
    def typeahead(self, workspace, params={}, **options): 
        """Retrieves objects in the workspace based on an auto-completion/typeahead
        search algorithm. This feature is meant to provide results quickly, so do
        not rely on this API to provide extremely accurate search results. The
        result set is limited to a single page of results with a maximum size,
        so you won't be able to fetch large numbers of results.

        Parameters
        ----------
        workspace : {Id} The workspace to fetch objects from.
        [params] : {Object} Parameters for the request
          - type : {Enum} The type of values the typeahead should return. You can choose from
          one of the following: custom_field, project, tag, task, and user.
          Note that unlike in the names of endpoints, the types listed here are
          in singular form (e.g. `task`). Using multiple types is not yet supported.
          - [query] : {String} The string that will be used to search for relevant objects. If an
          empty string is passed in, the API will currently return an empty
          result set.
          - [count] : {Number} The number of results to return. The default is `20` if this
          parameter is omitted, with a minimum of `1` and a maximum of `100`.
          If there are fewer results found than requested, all will be returned.
        """
        path = "/workspaces/%s/typeahead" % (workspace)
        return self.client.get_collection(path, params, **options)
        
    def add_user(self, workspace, params={}, **options): 
        """The user can be referenced by their globally unique user ID or their email address.
        Returns the full user record for the invited user.

        Parameters
        ----------
        workspace : {Id} The workspace or organization to invite the user to.
        [data] : {Object} Data for the request
          - user : {String} An identifier for the user. Can be one of an email address,
          the globally unique identifier for the user, or the keyword `me`
          to indicate the current user making the request.
        """
        path = "/workspaces/%s/addUser" % (workspace)
        return self.client.post(path, params, **options)
        
    def remove_user(self, workspace, params={}, **options): 
        """The user making this call must be an admin in the workspace.
        Returns an empty data record.

        Parameters
        ----------
        workspace : {Id} The workspace or organization to invite the user to.
        [data] : {Object} Data for the request
          - user : {String} An identifier for the user. Can be one of an email address,
          the globally unique identifier for the user, or the keyword `me`
          to indicate the current user making the request.
        """
        path = "/workspaces/%s/removeUser" % (workspace)
        return self.client.post(path, params, **options)
        



from .gen.organization_exports import _OrganizationExports


class OrganizationExports(_OrganizationExports):
    """The :class:`OrganizationExports` object, which represents the resource
    of the same name in Asana's API.

    """




from .error import InvalidTokenError
import time

class PageIterator(object):
    """Generic page iterator class, used for collections and events"""

    def __init__(self, client, path, query, options):
        self.client = client
        self.path = path
        self.query = query
        self.options = client._merge_options(options, { 'full_payload': True })

        self.item_limit = float('inf') if self.options.get('item_limit', None) == None else self.options['item_limit']
        self.page_size = self.options['page_size']
        self.count = 0

        self.continuation = False

    def __getattr__(self, name):
        """Getter for the custom named 'continuation' object."""
        if name == self.CONTINUATION_KEY:
            return self.continuation
        raise AttributeError("%r object has no attribute %r" % (self.__class__, attr))

    def __iter__(self):
        """Iterator interface, self is an iterator"""
        return self

    def __next__(self):
        """Iterator interface, returns the next 'page'"""

        # Compute the limit from the page size, and remaining item limit
        self.options['limit'] = min(self.page_size, self.item_limit - self.count)
        # If there is no continuation value or computed limit is 0, we're done
        if self.continuation == None or self.options['limit'] == 0:
            raise StopIteration
        # First call to __next__
        elif self.continuation == False:
            result = self.get_initial()
        # Subsequent calls to __next__
        else:
            result = self.get_next()
        # Extract the continuation from the response
        self.continuation = result.get(self.CONTINUATION_KEY, None)
        # Get the data, update the count, return the data
        data = result.get('data', None)
        if data != None:
            self.count += len(data)
        return data

    def next(self):
        """Alias for __next__"""
        return self.__next__()

    def items(self):
        """Returns an iterator for each item in each page"""
        for page in self:
            for item in page:
                yield item
        raise StopIteration


class CollectionPageIterator(PageIterator):
    """Iterator that returns one page of a collection at a time"""

    CONTINUATION_KEY = 'next_page'

    def get_initial(self):
        return self.client.get(self.path, self.query, **self.options)

    def get_next(self):
        self.options['offset'] = self.continuation['offset']
        return self.client.get(self.path, self.query, **self.options)


class EventsPageIterator(PageIterator):
    """Iterator that returns the next page of events, polls until a non-empty page is returned"""

    CONTINUATION_KEY = 'sync'

    def get_initial(self):
        # If no sync token was provided, make a request to get one
        if 'sync' not in self.query:
            try:
                self.client.events.get(self.query, **self.options)
            except InvalidTokenError as e:
                # InvalidTokenError is expected to be thrown since we didn't provide a sync token
                self.continuation = e.sync
        else:
            self.continuation = self.query['sync']
        return self.get_next()

    def get_next(self):
        self.query['sync'] = self.continuation
        return self.client.events.get(self.query, **self.options)

    def __next__(self):
        # Override __next__ to continue polling until a non-empty page of events is returned
        while True:
            results = super(EventsPageIterator, self).__next__()
            if len(results) > 0:
                return results
            else:
                time.sleep(self.options['poll_interval'])




from .gen.attachments import _Attachments

class Attachments(_Attachments):
    """Attachments resource"""

    def create_on_task(self, task_id, file_content, file_name, file_content_type=None, **options):
        """Upload an attachment for a task. Accepts a file object or string, file name, and optional file Content-Type"""
        path = '/tasks/%d/attachments' % (task_id)
        return self.client.request('post', path, files=[('file', (file_name, file_content, file_content_type))], **options)



from . import (
    attachments,
    custom_field_settings,
    custom_fields,
    events,
    organization_exports,
    project_memberships,
    project_statuses,
    projects,
    sections,
    stories,
    tags,
    tasks,
    teams,
    users,
    webhooks,
    workspaces
)



from types import ModuleType
import json
import platform
import time
import string

import requests

from . import error, resources, session, __version__
from .page_iterator import CollectionPageIterator

# TODO: Consider using six here instead
try:
    import urllib.parse as urlparse
except ImportError:
    import urllib as urlparse

# Create a dict of resource classes
RESOURCE_CLASSES = {}
for name, module in resources.__dict__.items():
    classified_name = string.capwords(name, '_').replace('_', '')
    if isinstance(module, ModuleType) and classified_name in module.__dict__:
        RESOURCE_CLASSES[name] = module.__dict__[classified_name]

# Create a mapping of status codes to classes
STATUS_MAP = {}
for name, Klass in error.__dict__.items():
    if isinstance(Klass, type) and issubclass(Klass, error.AsanaError):
        STATUS_MAP[Klass().status] = Klass


class Client(object):
    """Asana client class"""

    DEFAULT_OPTIONS = {
        'base_url': 'https://app.asana.com/api/1.0',
        'item_limit': None,
        'page_size': 50,
        'poll_interval': 5,
        'max_retries': 5,
        'full_payload': False,
        'iterator_type': 'items'
    }

    RETRY_DELAY = 1.0
    RETRY_BACKOFF = 2.0

    CLIENT_OPTIONS = set(DEFAULT_OPTIONS.keys())
    QUERY_OPTIONS = {'limit', 'offset', 'sync'}
    REQUEST_OPTIONS = {'headers', 'params', 'data', 'files', 'verify',
                       'timeout'}
    API_OPTIONS = {'pretty', 'fields', 'expand', 'client_name'}

    ALL_OPTIONS = (
        CLIENT_OPTIONS | QUERY_OPTIONS | REQUEST_OPTIONS | API_OPTIONS)

    def __init__(self, session=None, auth=None, **options):
        """A :class:`Client` object for interacting with Asana's API.

        A Client object with session, optional auth handler, and options.

        """
        self.session = session or requests.Session()
        self.auth = auth
        self.headers = options.pop('headers', {})
        # merge the provided options (if any) with the global DEFAULTS
        self.options = _merge(self.DEFAULT_OPTIONS, options)
        # intializes each resource, injecting this client object into the
        # constructor
        for name, Klass in RESOURCE_CLASSES.items():
            setattr(self, name, Klass(self))

    def request(self, method, path, **options):
        """Dispatches a request to the Asana HTTP API"""
        options = self._merge_options(options)
        url = options['base_url'] + path
        retry_count = 0
        request_options = self._parse_request_options(options)
        self._add_version_header(request_options)
        while True:
            try:
                response = getattr(self.session, method)(
                    url, auth=self.auth, **request_options)
                if response.status_code in STATUS_MAP:
                    raise STATUS_MAP[response.status_code](response)
                elif 500 <= response.status_code < 600:
                    # Any unhandled 500 is a server error.
                    raise error.ServerError(response)
                else:
                    if options['full_payload']:
                        return response.json()
                    else:
                        return response.json()['data']
            except error.RetryableAsanaError as e:
                if retry_count < options['max_retries']:
                    self._handle_retryable_error(e, retry_count)
                    retry_count += 1
                else:
                    raise e

    def _handle_retryable_error(self, e, retry_count):
        """Sleep based on the type of :class:`RetryableAsanaError`"""
        if isinstance(e, error.RateLimitEnforcedError):
            time.sleep(e.retry_after)
        else:
            time.sleep(self.RETRY_DELAY * (self.RETRY_BACKOFF ** retry_count))

    def get(self, path, query, **options):
        """Parses GET request options and dispatches a request."""
        api_options = self._parse_api_options(options, query_string=True)
        query_options = self._parse_query_options(options)
        parameter_options = self._parse_parameter_options(options)

        # options in the query takes precendence
        query = _merge(query_options, api_options, parameter_options, query)
        return self.request('get', path, params=query, **options)

    def get_collection(self, path, query, **options):
        """Get a collection from a collection endpoint.

        Parses GET request options for a collection endpoint and dispatches a
        request.

        """
        options = self._merge_options(options)
        if options['iterator_type'] == 'items':
            return CollectionPageIterator(self, path, query, options).items()
        if options['iterator_type'] is None:
            return self.get(path, query, **options)
        raise Exception('Unknown value for "iterator_type" option: {}'.format(
            str(options['iterator_type'])))

    def post(self, path, data, **options):
        """Parses POST request options and dispatches a request."""
        parameter_options = self._parse_parameter_options(options)
        body = {
            # values in the data body takes precendence
            'data': _merge(parameter_options, data),
            'options': self._parse_api_options(options)
        }
        return self.request(
            'post', path, data=body,
            headers={'content-type': 'application/json'}, **options)

    def put(self, path, data, **options):
        """Parses PUT request options and dispatches a request."""
        parameter_options = self._parse_parameter_options(options)
        body = {
            # values in the data body takes precendence
            'data': _merge(parameter_options, data),
            'options': self._parse_api_options(options)
        }
        return self.request(
            'put', path, data=body,
            headers={'content-type': 'application/json'}, **options)

    def delete(self, path, data, **options):
        """Dispatches a DELETE request."""
        return self.request('delete', path, **options)

    def _merge_options(self, *objects):
        """Merge option objects with the client's object.

        Merges one or more options objects with client's options and returns a
        new options object.

        """
        return _merge(self.options, *objects)

    def _parse_query_options(self, options):
        """Select query string options out of the provided options object"""
        return self._select_options(options, self.QUERY_OPTIONS)

    def _parse_parameter_options(self, options):
        """Select all unknown options.

        Select all unknown options (not query string, API, or request
        options)

        """
        return self._select_options(options, self.ALL_OPTIONS, invert=True)

    def _parse_api_options(self, options, query_string=False):
        """Select API options out of the provided options object.

        Selects API string options out of the provided options object and
        formats for either request body (default) or query string.

        """
        api_options = self._select_options(options, self.API_OPTIONS)
        if query_string:
            # Prefix all options with "opt_"
            query_api_options = {}
            for key in api_options:
                # Transform list/tuples into comma separated list
                if isinstance(api_options[key], (list, tuple)):
                    query_api_options[
                        'opt_' + key] = ','.join(api_options[key])
                else:
                    query_api_options[
                        'opt_' + key] = api_options[key]
            return query_api_options
        else:
            return api_options

    def _parse_request_options(self, options):
        """Select request options out of the provided options object.


        Select and formats options to be passed to the 'requests' library's
        request methods.

        """
        request_options = self._select_options(options, self.REQUEST_OPTIONS)
        if 'params' in request_options:
            params = request_options['params']
            for key in params:
                if isinstance(params[key], bool):
                    params[key] = json.dumps(params[key])
        if 'data' in request_options:
            # remove empty 'options':
            if 'options' in request_options['data'] and (
                    len(request_options['data']['options']) == 0):
                del request_options['data']['options']
            # serialize 'data' to JSON, requests doesn't do this automatically:
            request_options['data'] = json.dumps(request_options['data'])

        headers = self.headers.copy()
        headers.update(request_options.get('headers', {}))
        request_options['headers'] = headers
        return request_options

    def _select_options(self, options, keys, invert=False):
        """Select the provided keys out of an options object.


        Selects the provided keys (or everything except the provided keys) out
        of an options object.

        """
        options = self._merge_options(options)
        result = {}
        for key in options:
            if (invert and key not in keys) or (not invert and key in keys):
                result[key] = options[key]
        return result

    def _add_version_header(self, options):
        """Add the client lib version header to the request."""
        headers = options.setdefault('headers', {})
        headers['X-Asana-Client-Lib'] = self._version_header()

    _cached_version_header = None

    def _version_header(self):
        """Generate the client version header to send on each request."""
        if not self._cached_version_header:
            self._cached_version_header = urlparse.urlencode(
                self._version_values())
        return self._cached_version_header

    def _version_values(self):
        """Generate the values to go in the client version header."""
        return {
            'language': 'Python',
            'version': __version__,
            'language_version': platform.python_version(),
            'os': platform.system(),
            'os_version': platform.release()
        }

    @classmethod
    def basic_auth(Klass, apiKey):
        """DEPRECATED: this is only present for backwards-compatibility.

        This will be removed in the future; for new apps, prefer the
        `access_token` method.

        Construct an Asana Client using a Personal Access Token as if it
        were an old (removed) Asana API Key.
        """
        return Klass(auth=requests.auth.HTTPBasicAuth(apiKey, ''))

    @classmethod
    def access_token(Klass, accessToken):
        """Construct an Asana Client with a Personal Access Token"""
        return Klass(
            session.AsanaOAuth2Session(token={'access_token': accessToken}))

    @classmethod
    def oauth(Klass, **kwargs):
        """Construct an Asana Client with Oauth credentials.

        Construct an Asana Client with OAuth credentials ('client_id' and
        'client_secret' or 'token').

        """
        return Klass(session.AsanaOAuth2Session(**kwargs))


def _merge(*objects):
    """Merge one or more objects into a new object"""
    result = {}
    [result.update(obj) for obj in objects]
    return result




class _Sections:
    """A _section_ is a subdivision of a project that groups tasks together. It can
    either be a header above a list of tasks in a list view or a column in a
    board view of a project.
    """

    def __init__(self, client=None):
        self.client = client
  
    def create_in_project(self, project, params={}, **options): 
        """Creates a new section in a project.
        
        Returns the full record of the newly created section.

        Parameters
        ----------
        project : {Id} The project to create the section in
        [data] : {Object} Data for the request
          - name : {String} The text to be displayed as the section name. This cannot be an empty string.
        """
        path = "/projects/%s/sections" % (project)
        return self.client.post(path, params, **options)
        
    def find_by_project(self, project, params={}, **options): 
        """Returns the compact records for all sections in the specified project.

        Parameters
        ----------
        project : {Id} The project to get sections from.
        [params] : {Object} Parameters for the request
        """
        path = "/projects/%s/sections" % (project)
        return self.client.get(path, params, **options)
        
    def find_by_id(self, section, params={}, **options): 
        """Returns the complete record for a single section.

        Parameters
        ----------
        section : {Id} The section to get.
        [params] : {Object} Parameters for the request
        """
        path = "/sections/%s" % (section)
        return self.client.get(path, params, **options)
        
    def update(self, section, params={}, **options): 
        """A specific, existing section can be updated by making a PUT request on
        the URL for that project. Only the fields provided in the `data` block
        will be updated; any unspecified fields will remain unchanged. (note that
        at this time, the only field that can be updated is the `name` field.)
        
        When using this method, it is best to specify only those fields you wish
        to change, or else you may overwrite changes made by another user since
        you last retrieved the task.
        
        Returns the complete updated section record.

        Parameters
        ----------
        section : {Id} The section to update.
        [data] : {Object} Data for the request
        """
        path = "/sections/%s" % (section)
        return self.client.put(path, params, **options)
        
    def delete(self, section, params={}, **options): 
        """A specific, existing section can be deleted by making a DELETE request
        on the URL for that section.
        
        Note that sections must be empty to be deleted.
        
        The last remaining section in a board view cannot be deleted.
        
        Returns an empty data block.

        Parameters
        ----------
        section : {Id} The section to delete.
        """
        path = "/sections/%s" % (section)
        return self.client.delete(path, params, **options)
        
    def insert_in_project(self, project, params={}, **options): 
        """Move sections relative to each other in a board view. One of
        `before_section` or `after_section` is required.
        
        Sections cannot be moved between projects.
        
        At this point in time, moving sections is not supported in list views, only board views.
        
        Returns an empty data block.

        Parameters
        ----------
        project : {Id} The project in which to reorder the given section
        [data] : {Object} Data for the request
          - section : {Id} The section to reorder
          - [before_section] : {Id} Insert the given section immediately before the section specified by this parameter.
          - [after_section] : {Id} Insert the given section immediately after the section specified by this parameter.
        """
        path = "/projects/%s/sections/insert" % (project)
        return self.client.post(path, params, **options)
        




from .gen.tasks import _Tasks

class Tasks(_Tasks):
    """Tasks resource"""
    def set_parent(self, task_id, params={}, **options):
        """Changes the parent of a task. Each task may only be a subtask of a single
        parent, or no parent task at all. Returns an empty data block.

        Parameters
        ----------
        task : {Id} Globally unique identifier for the task.
        [data] : {Object} Data for the request
          - parent : {Id} The new parent of the task, or `null` for no parent.
        """
        path = '/tasks/%s/setParent' % (task_id)
        return self.client.post(path, params, **options)




class _ProjectStatuses:
    """A _project status_ is an update on the progress of a particular project, and is sent out to all project
    followers when created. These updates include both text describing the update and a color code intended to
    represent the overall state of the project: "green" for projects that are on track, "yellow" for projects
    at risk, and "red" for projects that are behind.
    
    Project statuses can be created and deleted, but not modified.
    """

    def __init__(self, client=None):
        self.client = client
  
    def create_in_project(self, project, params={}, **options): 
        """Creates a new status update on the project.
        
        Returns the full record of the newly created project status update.

        Parameters
        ----------
        project : {Id} The project on which to create a status update.
        [data] : {Object} Data for the request
          - text : {String} The text of the project status update.
          - color : {String} The color to associate with the status update. Must be one of `"red"`, `"yellow"`, or `"green"`.
        """
        path = "/projects/%s/project_statuses" % (project)
        return self.client.post(path, params, **options)
        
    def find_by_project(self, project, params={}, **options): 
        """Returns the compact project status update records for all updates on the project.

        Parameters
        ----------
        project : {Id} The project to find status updates for.
        [params] : {Object} Parameters for the request
        """
        path = "/projects/%s/project_statuses" % (project)
        return self.client.get_collection(path, params, **options)
        
    def find_by_id(self, project_status, params={}, **options): 
        """Returns the complete record for a single status update.

        Parameters
        ----------
        project-status : {Id} The project status update to get.
        [params] : {Object} Parameters for the request
        """
        path = "/project_statuses/%s" % (project_status)
        return self.client.get(path, params, **options)
        
    def delete(self, project_status, params={}, **options): 
        """Deletes a specific, existing project status update.
        
        Returns an empty data record.

        Parameters
        ----------
        project-status : {Id} The project status update to delete.
        """
        path = "/project_statuses/%s" % (project_status)
        return self.client.delete(path, params, **options)
        




class _Stories:
    """A _story_ represents an activity associated with an object in the Asana
    system. Stories are generated by the system whenever users take actions such
    as creating or assigning tasks, or moving tasks between projects. _Comments_
    are also a form of user-generated story.
    
    Stories are a form of history in the system, and as such they are read-only.
    Once generated, it is not possible to modify a story.
    """

    def __init__(self, client=None):
        self.client = client
  
    def find_by_task(self, task, params={}, **options): 
        """Returns the compact records for all stories on the task.

        Parameters
        ----------
        task : {Id} Globally unique identifier for the task.
        [params] : {Object} Parameters for the request
        """
        path = "/tasks/%s/stories" % (task)
        return self.client.get_collection(path, params, **options)
        
    def find_by_id(self, story, params={}, **options): 
        """Returns the full record for a single story.

        Parameters
        ----------
        story : {Id} Globally unique identifier for the story.
        [params] : {Object} Parameters for the request
        """
        path = "/stories/%s" % (story)
        return self.client.get(path, params, **options)
        
    def create_on_task(self, task, params={}, **options): 
        """Adds a comment to a task. The comment will be authored by the
        currently authenticated user, and timestamped when the server receives
        the request.
        
        Returns the full record for the new story added to the task.

        Parameters
        ----------
        task : {Id} Globally unique identifier for the task.
        [data] : {Object} Data for the request
          - text : {String} The plain text of the comment to add.
        """
        path = "/tasks/%s/stories" % (task)
        return self.client.post(path, params, **options)
        
    def update(self, story, params={}, **options): 
        """Updates the story and returns the full record for the updated story.
        Only comment stories can have their text updated, and only comment stories and
        attachment stories can be pinned. Only one of `text` and `html_text` can be specified.

        Parameters
        ----------
        story : {Id} Globally unique identifier for the story.
        [data] : {Object} Data for the request
          - [text] : {String} The plain text with which to update the comment.
          - [html_text] : {String} The rich text with which to update the comment.
          - [is_pinned] : {Boolean} Whether the story should be pinned on the resource.
        """
        path = "/stories/%s" % (story)
        return self.client.put(path, params, **options)
        
    def delete(self, story, params={}, **options): 
        """Deletes a story. A user can only delete stories they have created. Returns an empty data record.

        Parameters
        ----------
        story : {Id} Globally unique identifier for the story.
        """
        path = "/stories/%s" % (story)
        return self.client.delete(path, params, **options)
        



class AsanaError(Exception):
    """Base Asana error class"""
    def __init__(self, message=None, status=None, response=None):
        try:
            messages = [
                error['message'] for error in response.json()['errors']]
            message = message + ': ' + '; '.join(messages)
        except Exception:
            pass

        super(AsanaError, self).__init__(message)

        self.status = status
        self.response = response
        self.message = message


class InvalidRequestError(AsanaError):
    def __init__(self, response=None):
        super(InvalidRequestError, self).__init__(
            message='Invalid Request',
            status=400,
            response=response
        )


class NoAuthorizationError(AsanaError):
    def __init__(self, response=None):
        super(NoAuthorizationError, self).__init__(
            message='No Authorization',
            status=401,
            response=response
        )


class PremiumOnlyError(AsanaError):
    def __init__(self, response=None):
        super(PremiumOnlyError, self).__init__(
            message='Payment Required',
            status=402,
            response=response
        )


class ForbiddenError(AsanaError):
    def __init__(self, response=None):
        super(ForbiddenError, self).__init__(
            message='Forbidden',
            status=403,
            response=response
        )


class NotFoundError(AsanaError):
    def __init__(self, response=None):
        super(NotFoundError, self).__init__(
            message='Not Found',
            status=404,
            response=response
        )


class InvalidTokenError(AsanaError):
    def __init__(self, response=None):
        super(InvalidTokenError, self).__init__(
            message='Sync token invalid or too old',
            status=412,
            response=response
        )
        self.sync = response is not None and response.json()['sync']


class RetryableAsanaError(AsanaError):
    """Base class for retryable errors.

    Base class for errors which should trigger a retry (if configured to do
    so).

    """
    def __init__(self, message=None, status=None, response=None):
        super(RetryableAsanaError, self).__init__(
            message=message, status=status, response=response)


class RateLimitEnforcedError(RetryableAsanaError):
    def __init__(self, response=None):
        super(RateLimitEnforcedError, self).__init__(
            message='Rate Limit Enforced',
            status=429,
            response=response
        )
        self.retry_after = (
            response is not None and float(response.headers['Retry-After']))


class ServerError(RetryableAsanaError):
    def __init__(self, response=None):
        status = 500
        if response:
            status = response.status
        super(ServerError, self).__init__(
            message='Server Error',
            status=status,
            response=response
        )




class _Events:
    """An _event_ is an object representing a change to a resource that was observed
    by an event subscription.
    
    In general, requesting events on a resource is faster and subject to higher
    rate limits than requesting the resource itself. Additionally, change events
    bubble up - listening to events on a project would include when stories are
    added to tasks in the project, even on subtasks.
    
    Establish an initial sync token by making a request with no sync token.
    The response will be a `412` error - the same as if the sync token had
    expired.
    
    Subsequent requests should always provide the sync token from the immediately
    preceding call.
    
    Sync tokens may not be valid if you attempt to go 'backward' in the history
    by requesting previous tokens, though re-requesting the current sync token
    is generally safe, and will always return the same results.
    
    When you receive a `412 Precondition Failed` error, it means that the
    sync token is either invalid or expired. If you are attempting to keep a set
    of data in sync, this signals you may need to re-crawl the data.
    
    Sync tokens always expire after 24 hours, but may expire sooner, depending on
    load on the service.
    """

    def __init__(self, client=None):
        self.client = client
  




class _Tags:
    """A _tag_ is a label that can be attached to any task in Asana. It exists in a
    single workspace or organization.
    
    Tags have some metadata associated with them, but it is possible that we will
    simplify them in the future so it is not encouraged to rely too heavily on it.
    Unlike projects, tags do not provide any ordering on the tasks they
    are associated with.
    """

    def __init__(self, client=None):
        self.client = client
  
    def create(self, params={}, **options): 
        """Creates a new tag in a workspace or organization.
        
        Every tag is required to be created in a specific workspace or
        organization, and this cannot be changed once set. Note that you can use
        the `workspace` parameter regardless of whether or not it is an
        organization.
        
        Returns the full record of the newly created tag.

        Parameters
        ----------
        [data] : {Object} Data for the request
          - workspace : {Id} The workspace or organization to create the tag in.
        """
        return self.client.post("/tags", params, **options)
        
    def create_in_workspace(self, workspace, params={}, **options): 
        """Creates a new tag in a workspace or organization.
        
        Every tag is required to be created in a specific workspace or
        organization, and this cannot be changed once set. Note that you can use
        the `workspace` parameter regardless of whether or not it is an
        organization.
        
        Returns the full record of the newly created tag.

        Parameters
        ----------
        workspace : {Id} The workspace or organization to create the tag in.
        [data] : {Object} Data for the request
        """
        path = "/workspaces/%s/tags" % (workspace)
        return self.client.post(path, params, **options)
        
    def find_by_id(self, tag, params={}, **options): 
        """Returns the complete tag record for a single tag.

        Parameters
        ----------
        tag : {Id} The tag to get.
        [params] : {Object} Parameters for the request
        """
        path = "/tags/%s" % (tag)
        return self.client.get(path, params, **options)
        
    def update(self, tag, params={}, **options): 
        """Updates the properties of a tag. Only the fields provided in the `data`
        block will be updated; any unspecified fields will remain unchanged.
        
        When using this method, it is best to specify only those fields you wish
        to change, or else you may overwrite changes made by another user since
        you last retrieved the task.
        
        Returns the complete updated tag record.

        Parameters
        ----------
        tag : {Id} The tag to update.
        [data] : {Object} Data for the request
        """
        path = "/tags/%s" % (tag)
        return self.client.put(path, params, **options)
        
    def delete(self, tag, params={}, **options): 
        """A specific, existing tag can be deleted by making a DELETE request
        on the URL for that tag.
        
        Returns an empty data record.

        Parameters
        ----------
        tag : {Id} The tag to delete.
        """
        path = "/tags/%s" % (tag)
        return self.client.delete(path, params, **options)
        
    def find_all(self, params={}, **options): 
        """Returns the compact tag records for some filtered set of tags.
        Use one or more of the parameters provided to filter the tags returned.

        Parameters
        ----------
        [params] : {Object} Parameters for the request
          - [workspace] : {Id} The workspace or organization to filter tags on.
          - [team] : {Id} The team to filter tags on.
          - [archived] : {Boolean} Only return tags whose `archived` field takes on the value of
          this parameter.
        """
        return self.client.get_collection("/tags", params, **options)
        
    def find_by_workspace(self, workspace, params={}, **options): 
        """Returns the compact tag records for all tags in the workspace.

        Parameters
        ----------
        workspace : {Id} The workspace or organization to find tags in.
        [params] : {Object} Parameters for the request
        """
        path = "/workspaces/%s/tags" % (workspace)
        return self.client.get_collection(path, params, **options)
        
    def get_tasks_with_tag(self, tag, params={}, **options): 
        """Returns the compact task records for all tasks with the given tag.
        Tasks can have more than one tag at a time.

        Parameters
        ----------
        tag : {Id} The tag to fetch tasks from.
        [params] : {Object} Parameters for the request
        """
        path = "/tags/%s/tasks" % (tag)
        return self.client.get_collection(path, params, **options)
        




class _OrganizationExports:
    """An _organization_export_ object represents a request to export the complete data of an Organization
    in JSON format.
    
    To export an Organization using this API:
    
    * Create an `organization_export` [request](#create) and store the id that is returned.\
    * Request the `organization_export` every few minutes, until the `state` field contains 'finished'.\
    * Download the file located at the URL in the `download_url` field.
    
    Exports can take a long time, from several minutes to a few hours for large Organizations.
    
    **Note:** These endpoints are only available to [Service Accounts](/guide/help/premium/service-accounts)
    of an [Enterprise](/enterprise) Organization.
    """

    def __init__(self, client=None):
        self.client = client
  
    def find_by_id(self, organization_export, params={}, **options): 
        """Returns details of a previously-requested Organization export.

        Parameters
        ----------
        organization_export : {Id} Globally unique identifier for the Organization export.
        [params] : {Object} Parameters for the request
        """
        path = "/organization_exports/%s" % (organization_export)
        return self.client.get(path, params, **options)
        
    def create(self, params={}, **options): 
        """This method creates a request to export an Organization. Asana will complete the export at some
        point after you create the request.

        Parameters
        ----------
        [data] : {Object} Data for the request
          - organization : {Id} Globally unique identifier for the workspace or organization.
        """
        return self.client.post("/organization_exports", params, **options)
        




class _Teams:
    """A _team_ is used to group related projects and people together within an
    organization. Each project in an organization is associated with a team.
    """

    def __init__(self, client=None):
        self.client = client
  
    def find_by_id(self, team, params={}, **options): 
        """Returns the full record for a single team.

        Parameters
        ----------
        team : {Id} Globally unique identifier for the team.
        [params] : {Object} Parameters for the request
        """
        path = "/teams/%s" % (team)
        return self.client.get(path, params, **options)
        
    def find_by_organization(self, organization, params={}, **options): 
        """Returns the compact records for all teams in the organization visible to
        the authorized user.

        Parameters
        ----------
        organization : {Id} Globally unique identifier for the workspace or organization.
        [params] : {Object} Parameters for the request
        """
        path = "/organizations/%s/teams" % (organization)
        return self.client.get_collection(path, params, **options)
        
    def find_by_user(self, user, params={}, **options): 
        """Returns the compact records for all teams to which user is assigned.

        Parameters
        ----------
        user : {String} An identifier for the user. Can be one of an email address,
        the globally unique identifier for the user, or the keyword `me`
        to indicate the current user making the request.
        [params] : {Object} Parameters for the request
          - [organization] : {Id} The workspace or organization to filter teams on.
        """
        path = "/users/%s/teams" % (user)
        return self.client.get_collection(path, params, **options)
        
    def users(self, team, params={}, **options): 
        """Returns the compact records for all users that are members of the team.

        Parameters
        ----------
        team : {Id} Globally unique identifier for the team.
        [params] : {Object} Parameters for the request
        """
        path = "/teams/%s/users" % (team)
        return self.client.get_collection(path, params, **options)
        
    def add_user(self, team, params={}, **options): 
        """The user making this call must be a member of the team in order to add others.
        The user to add must exist in the same organization as the team in order to be added.
        The user to add can be referenced by their globally unique user ID or their email address.
        Returns the full user record for the added user.

        Parameters
        ----------
        team : {Id} Globally unique identifier for the team.
        [data] : {Object} Data for the request
          - user : {String} An identifier for the user. Can be one of an email address,
          the globally unique identifier for the user, or the keyword `me`
          to indicate the current user making the request.
        """
        path = "/teams/%s/addUser" % (team)
        return self.client.post(path, params, **options)
        
    def remove_user(self, team, params={}, **options): 
        """The user to remove can be referenced by their globally unique user ID or their email address.
        Removes the user from the specified team. Returns an empty data record.

        Parameters
        ----------
        team : {Id} Globally unique identifier for the team.
        [data] : {Object} Data for the request
          - user : {String} An identifier for the user. Can be one of an email address,
          the globally unique identifier for the user, or the keyword `me`
          to indicate the current user making the request.
        """
        path = "/teams/%s/removeUser" % (team)
        return self.client.post(path, params, **options)
        




class _Tasks:
    """The _task_ is the basic object around which many operations in Asana are
    centered. In the Asana application, multiple tasks populate the middle pane
    according to some view parameters, and the set of selected tasks determines
    the more detailed information presented in the details pane.
    """

    def __init__(self, client=None):
        self.client = client
  
    def create(self, params={}, **options): 
        """Creating a new task is as easy as POSTing to the `/tasks` endpoint
        with a data block containing the fields you'd like to set on the task.
        Any unspecified fields will take on default values.
        
        Every task is required to be created in a specific workspace, and this
        workspace cannot be changed once set. The workspace need not be set
        explicitly if you specify `projects` or a `parent` task instead.
        
        `projects` can be a comma separated list of projects, or just a single
        project the task should belong to.

        Parameters
        ----------
        [data] : {Object} Data for the request
          - [workspace] : {Id} The workspace to create a task in.
        """
        return self.client.post("/tasks", params, **options)
        
    def create_in_workspace(self, workspace, params={}, **options): 
        """Creating a new task is as easy as POSTing to the `/tasks` endpoint
        with a data block containing the fields you'd like to set on the task.
        Any unspecified fields will take on default values.
        
        Every task is required to be created in a specific workspace, and this
        workspace cannot be changed once set. The workspace need not be set
        explicitly if you specify a `project` or a `parent` task instead.

        Parameters
        ----------
        workspace : {Id} The workspace to create a task in.
        [data] : {Object} Data for the request
        """
        path = "/workspaces/%s/tasks" % (workspace)
        return self.client.post(path, params, **options)
        
    def find_by_id(self, task, params={}, **options): 
        """Returns the complete task record for a single task.

        Parameters
        ----------
        task : {Id} The task to get.
        [params] : {Object} Parameters for the request
        """
        path = "/tasks/%s" % (task)
        return self.client.get(path, params, **options)
        
    def update(self, task, params={}, **options): 
        """A specific, existing task can be updated by making a PUT request on the
        URL for that task. Only the fields provided in the `data` block will be
        updated; any unspecified fields will remain unchanged.
        
        When using this method, it is best to specify only those fields you wish
        to change, or else you may overwrite changes made by another user since
        you last retrieved the task.
        
        Returns the complete updated task record.

        Parameters
        ----------
        task : {Id} The task to update.
        [data] : {Object} Data for the request
        """
        path = "/tasks/%s" % (task)
        return self.client.put(path, params, **options)
        
    def delete(self, task, params={}, **options): 
        """A specific, existing task can be deleted by making a DELETE request on the
        URL for that task. Deleted tasks go into the "trash" of the user making
        the delete request. Tasks can be recovered from the trash within a period
        of 30 days; afterward they are completely removed from the system.
        
        Returns an empty data record.

        Parameters
        ----------
        task : {Id} The task to delete.
        """
        path = "/tasks/%s" % (task)
        return self.client.delete(path, params, **options)
        
    def find_by_project(self, project_id, params={}, **options): 
        """Returns the compact task records for all tasks within the given project,
        ordered by their priority within the project.

        Parameters
        ----------
        projectId : {Id} The project in which to search for tasks.
        [params] : {Object} Parameters for the request
        """
        path = "/projects/%s/tasks" % (project_id)
        return self.client.get_collection(path, params, **options)
        
    def find_by_tag(self, tag, params={}, **options): 
        """Returns the compact task records for all tasks with the given tag.

        Parameters
        ----------
        tag : {Id} The tag in which to search for tasks.
        [params] : {Object} Parameters for the request
        """
        path = "/tags/%s/tasks" % (tag)
        return self.client.get_collection(path, params, **options)
        
    def find_by_section(self, section, params={}, **options): 
        """<b>Board view only:</b> Returns the compact section records for all tasks within the given section.

        Parameters
        ----------
        section : {Id} The section in which to search for tasks.
        [params] : {Object} Parameters for the request
        """
        path = "/sections/%s/tasks" % (section)
        return self.client.get_collection(path, params, **options)
        
    def find_all(self, params={}, **options): 
        """Returns the compact task records for some filtered set of tasks. Use one
        or more of the parameters provided to filter the tasks returned. You must
        specify a `project` or `tag` if you do not specify `assignee` and `workspace`.

        Parameters
        ----------
        [params] : {Object} Parameters for the request
          - [assignee] : {String} The assignee to filter tasks on.
          - [project] : {Id} The project to filter tasks on.
          - [section] : {Id} The section to filter tasks on.
          - [workspace] : {Id} The workspace or organization to filter tasks on.
          - [completed_since] : {String} Only return tasks that are either incomplete or that have been
          completed since this time.
          - [modified_since] : {String} Only return tasks that have been modified since the given time.
        """
        return self.client.get_collection("/tasks", params, **options)
        
    def search_in_workspace(self, workspace, params={}, **options): 
        """The search endpoint allows you to build complex queries to find and fetch exactly the data you need from Asana. For a more comprehensive description of all the query parameters and limitations of this endpoint, see our [long-form documentation](/developers/documentation/getting-started/search-api) for this feature.

        Parameters
        ----------
        workspace : {Id} The workspace or organization in which to search for tasks.
        [params] : {Object} Parameters for the request
        """
        path = "/workspaces/%s/tasks/search" % (workspace)
        return self.client.get_collection(path, params, **options)
        
    def dependencies(self, task, params={}, **options): 
        """Returns the compact representations of all of the dependencies of a task.

        Parameters
        ----------
        task : {Id} The task to get dependencies on.
        [params] : {Object} Parameters for the request
        """
        path = "/tasks/%s/dependencies" % (task)
        return self.client.get(path, params, **options)
        
    def dependents(self, task, params={}, **options): 
        """Returns the compact representations of all of the dependents of a task.

        Parameters
        ----------
        task : {Id} The task to get dependents on.
        [params] : {Object} Parameters for the request
        """
        path = "/tasks/%s/dependents" % (task)
        return self.client.get(path, params, **options)
        
    def add_dependencies(self, task, params={}, **options): 
        """Marks a set of tasks as dependencies of this task, if they are not
        already dependencies. *A task can have at most 15 dependencies.*

        Parameters
        ----------
        task : {Id} The task to add dependencies to.
        [data] : {Object} Data for the request
          - dependencies : {Array} An array of task IDs that this task should depend on.
        """
        path = "/tasks/%s/addDependencies" % (task)
        return self.client.post(path, params, **options)
        
    def add_dependents(self, task, params={}, **options): 
        """Marks a set of tasks as dependents of this task, if they are not already
        dependents. *A task can have at most 30 dependents.*

        Parameters
        ----------
        task : {Id} The task to add dependents to.
        [data] : {Object} Data for the request
          - dependents : {Array} An array of task IDs that should depend on this task.
        """
        path = "/tasks/%s/addDependents" % (task)
        return self.client.post(path, params, **options)
        
    def remove_dependencies(self, task, params={}, **options): 
        """Unlinks a set of dependencies from this task.

        Parameters
        ----------
        task : {Id} The task to remove dependencies from.
        [data] : {Object} Data for the request
          - dependencies : {Array} An array of task IDs to remove as dependencies.
        """
        path = "/tasks/%s/removeDependencies" % (task)
        return self.client.post(path, params, **options)
        
    def remove_dependents(self, task, params={}, **options): 
        """Unlinks a set of dependents from this task.

        Parameters
        ----------
        task : {Id} The task to remove dependents from.
        [data] : {Object} Data for the request
          - dependents : {Array} An array of task IDs to remove as dependents.
        """
        path = "/tasks/%s/removeDependents" % (task)
        return self.client.post(path, params, **options)
        
    def add_followers(self, task, params={}, **options): 
        """Adds each of the specified followers to the task, if they are not already
        following. Returns the complete, updated record for the affected task.

        Parameters
        ----------
        task : {Id} The task to add followers to.
        [data] : {Object} Data for the request
          - followers : {Array} An array of followers to add to the task.
        """
        path = "/tasks/%s/addFollowers" % (task)
        return self.client.post(path, params, **options)
        
    def remove_followers(self, task, params={}, **options): 
        """Removes each of the specified followers from the task if they are
        following. Returns the complete, updated record for the affected task.

        Parameters
        ----------
        task : {Id} The task to remove followers from.
        [data] : {Object} Data for the request
          - followers : {Array} An array of followers to remove from the task.
        """
        path = "/tasks/%s/removeFollowers" % (task)
        return self.client.post(path, params, **options)
        
    def projects(self, task, params={}, **options): 
        """Returns a compact representation of all of the projects the task is in.

        Parameters
        ----------
        task : {Id} The task to get projects on.
        [params] : {Object} Parameters for the request
        """
        path = "/tasks/%s/projects" % (task)
        return self.client.get_collection(path, params, **options)
        
    def add_project(self, task, params={}, **options): 
        """Adds the task to the specified project, in the optional location
        specified. If no location arguments are given, the task will be added to
        the end of the project.
        
        `addProject` can also be used to reorder a task within a project or section that
        already contains it.
        
        At most one of `insert_before`, `insert_after`, or `section` should be
        specified. Inserting into a section in an non-order-dependent way can be
        done by specifying `section`, otherwise, to insert within a section in a
        particular place, specify `insert_before` or `insert_after` and a task
        within the section to anchor the position of this task.
        
        Returns an empty data block.

        Parameters
        ----------
        task : {Id} The task to add to a project.
        [data] : {Object} Data for the request
          - project : {Id} The project to add the task to.
          - [insert_after] : {Id} A task in the project to insert the task after, or `null` to
          insert at the beginning of the list.
          - [insert_before] : {Id} A task in the project to insert the task before, or `null` to
          insert at the end of the list.
          - [section] : {Id} A section in the project to insert the task into. The task will be
          inserted at the bottom of the section.
        """
        path = "/tasks/%s/addProject" % (task)
        return self.client.post(path, params, **options)
        
    def remove_project(self, task, params={}, **options): 
        """Removes the task from the specified project. The task will still exist
        in the system, but it will not be in the project anymore.
        
        Returns an empty data block.

        Parameters
        ----------
        task : {Id} The task to remove from a project.
        [data] : {Object} Data for the request
          - project : {Id} The project to remove the task from.
        """
        path = "/tasks/%s/removeProject" % (task)
        return self.client.post(path, params, **options)
        
    def tags(self, task, params={}, **options): 
        """Returns a compact representation of all of the tags the task has.

        Parameters
        ----------
        task : {Id} The task to get tags on.
        [params] : {Object} Parameters for the request
        """
        path = "/tasks/%s/tags" % (task)
        return self.client.get_collection(path, params, **options)
        
    def add_tag(self, task, params={}, **options): 
        """Adds a tag to a task. Returns an empty data block.

        Parameters
        ----------
        task : {Id} The task to add a tag to.
        [data] : {Object} Data for the request
          - tag : {Id} The tag to add to the task.
        """
        path = "/tasks/%s/addTag" % (task)
        return self.client.post(path, params, **options)
        
    def remove_tag(self, task, params={}, **options): 
        """Removes a tag from the task. Returns an empty data block.

        Parameters
        ----------
        task : {Id} The task to remove a tag from.
        [data] : {Object} Data for the request
          - tag : {Id} The tag to remove from the task.
        """
        path = "/tasks/%s/removeTag" % (task)
        return self.client.post(path, params, **options)
        
    def subtasks(self, task, params={}, **options): 
        """Returns a compact representation of all of the subtasks of a task.

        Parameters
        ----------
        task : {Id} The task to get the subtasks of.
        [params] : {Object} Parameters for the request
        """
        path = "/tasks/%s/subtasks" % (task)
        return self.client.get_collection(path, params, **options)
        
    def add_subtask(self, task, params={}, **options): 
        """Creates a new subtask and adds it to the parent task. Returns the full record
        for the newly created subtask.

        Parameters
        ----------
        task : {Id} The task to add a subtask to.
        [data] : {Object} Data for the request
        """
        path = "/tasks/%s/subtasks" % (task)
        return self.client.post(path, params, **options)
        
    def stories(self, task, params={}, **options): 
        """Returns a compact representation of all of the stories on the task.

        Parameters
        ----------
        task : {Id} The task containing the stories to get.
        [params] : {Object} Parameters for the request
        """
        path = "/tasks/%s/stories" % (task)
        return self.client.get_collection(path, params, **options)
        
    def add_comment(self, task, params={}, **options): 
        """Adds a comment to a task. The comment will be authored by the
        currently authenticated user, and timestamped when the server receives
        the request.
        
        Returns the full record for the new story added to the task.

        Parameters
        ----------
        task : {Id} Globally unique identifier for the task.
        [data] : {Object} Data for the request
          - text : {String} The plain text of the comment to add.
        """
        path = "/tasks/%s/stories" % (task)
        return self.client.post(path, params, **options)
        



#!/usr/bin/env python

import sys
import os
from setuptools import setup, find_packages

assert sys.version_info >= (2, 7), 'We only support Python 2.7+'

sys.path.insert(0, os.path.join(os.path.dirname(__file__), 'asana'))

setup(
    name='asana',
    version='0.8.0',
    description='Asana API client',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
    install_requires=[
        'requests >= 2.14, == 2.*',
        'requests_oauthlib >= 0.8.0, == 0.8.*',
        'six >= 1.10, == 1.*'
    ],
    author='Asana, Inc',
    # author_email='',
    url='http://github.com/asana/python-asana',
    packages=find_packages(exclude=('tests', 'examples')),
    keywords='asana',
    zip_safe=True,
    test_suite='tests')



import requests
import requests_oauthlib

class AsanaOAuth2Session(requests_oauthlib.OAuth2Session):
    """Session class for making OAuth authenticated requests to Asana's API"""

    authorize_url = 'https://app.asana.com/-/oauth_authorize'
    token_url = 'https://app.asana.com/-/oauth_token'

    def __init__(self, client_secret=None, **kwargs):
        super(AsanaOAuth2Session, self).__init__(**kwargs)
        self.client_secret = client_secret

    def authorization_url(self):
        """Get a URL (and 'state' token) to redirect the user to for authentication"""
        return super(AsanaOAuth2Session, self).authorization_url(self.authorize_url)

    def fetch_token(self, **kwargs):
        """Exchange a code (and 'state' token) for a bearer token"""
        return super(AsanaOAuth2Session, self).fetch_token(self.token_url, client_secret=self.client_secret, **kwargs)




class _CustomFieldSettings:
    """Custom fields are attached to a particular project with the Custom
    Field Settings resource. This resource both represents the many-to-many join
    of the Custom Field and Project as well as stores information that is relevant to that
    particular pairing; for instance, the `is_important` property determines
    some possible application-specific handling of that custom field (see below)
    """

    def __init__(self, client=None):
        self.client = client
  
    def find_by_project(self, project, params={}, **options): 
        """Returns a list of all of the custom fields settings on a project, in compact form. Note that, as in all queries to collections which return compact representation, `opt_fields` and `opt_expand` can be used to include more data than is returned in the compact representation. See the getting started guide on [input/output options](/developers/documentation/getting-started/input-output-options) for more information.

        Parameters
        ----------
        project : {Id} The ID of the project for which to list custom field settings
        [params] : {Object} Parameters for the request
        """
        path = "/projects/%s/custom_field_settings" % (project)
        return self.client.get_collection(path, params, **options)
        




class _ProjectMemberships:
    """With the introduction of "comment-only" projects in Asana, a user's membership
    in a project comes with associated permissions. These permissions (whether a
    user has full access to the project or comment-only access) are accessible
    through the project memberships endpoints described here.
    """

    def __init__(self, client=None):
        self.client = client
  
    def find_by_project(self, project, params={}, **options): 
        """Returns the compact project membership records for the project.

        Parameters
        ----------
        project : {Id} The project for which to fetch memberships.
        [params] : {Object} Parameters for the request
          - [user] : {String} If present, the user to filter the memberships to.
        """
        path = "/projects/%s/project_memberships" % (project)
        return self.client.get_collection(path, params, **options)
        
    def find_by_id(self, project_membership, params={}, **options): 
        """Returns the project membership record.

        Parameters
        ----------
        project_membership : {Id} Globally unique identifier for the project membership.
        [params] : {Object} Parameters for the request
        """
        path = "/project_memberships/%s" % (project_membership)
        return self.client.get(path, params, **options)
        
